import abc
import copy
import functools
import hashlib
import io
import pathlib

import numpy as np
import qpimage


class SeriesData(object):
    """Series data file format base class

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the experimental data file.
    meta_data: dict
        Dictionary containing meta data.
        see :py:class:`qpimage.META_KEYS`.
    as_type: str
        Defines the data type that the input data is casted to.
        The default is "float32" which saves memory. If high
        numerical accuracy is required (does not apply for a
        simple 2D phase analysis), set this to double precision
        ("float64").
    """
    __meta__ = abc.ABCMeta
    is_series = True
    priority = 0  # decrease to get higher priority

    def __init__(self, path, meta_data=None, holo_kw=None, as_type="float32"):
        #: Enforced dtype via keyword arguments
        if holo_kw is None:
            holo_kw = {}
        if meta_data is None:
            meta_data = {}
        self.as_type = as_type
        if isinstance(path, io.IOBase):
            # io.IOBase
            self.path = path
        else:
            #: pathlib.Path to data file or io.IOBase
            self.path = pathlib.Path(path).resolve()

        # check for valid metadata keys
        for key in meta_data:
            if key not in qpimage.meta.META_KEYS:
                msg = "Invalid metadata key `{}`!".format(key) \
                      + "Valid keys: {}".format(sorted(qpimage.meta.META_KEYS))
                raise ValueError(msg)
        #: Enforced metadata via keyword arguments
        self.meta_data = copy.copy(meta_data)
        #: Hologram retrieval; keyword arguments for
        #: :func:`qpimage.holo.get_field`.
        self.holo_kw = holo_kw
        self._bgdata = []
        #: Unique string that identifies the background data that
        #: was set using `set_bg`.
        self.background_identifier = None
        #: the file format name
        self.format = self.__class__.__name__

    def __repr__(self):
        rep = f"QPFormat '{self.format}'" \
              + f", {len(self)} image(s)" \
              + f"at {hex(id(self))}" \
              + "\nfile: {self.path}"

        meta = []
        if "wavelength" in self.meta_data:
            wl = self.meta_data["wavelength"]
            if 2000e-9 > wl > 10e-9:
                # convenience for light microscopy
                meta.append(f"λ={wl * 1e9:.1f}nm")
            else:
                meta.append(f"λ={wl:.2e}m")
        if "pixel size" in self.meta_data:
            pxm = self.meta_data["pixel size"]
            if 1e-3 > pxm > 1e-8:
                # convenience for light microscopy
                meta.append(f"1px={pxm * 1e6}µm")
            else:
                meta.append(f"1px={pxm}m")
        rep = ", ".join([rep] + meta)
        return rep

    @abc.abstractmethod
    def __len__(self):
        """Return number of samples of a data set"""

    def _compute_bgid(self, bg=None):
        """Return a unique identifier for the background data"""
        if bg is None:
            bg = self._bgdata
        if isinstance(bg, qpimage.QPImage):
            # Single QPImage
            if "identifier" in bg:
                return bg["identifier"]
            else:
                data = [bg.amp, bg.pha]
                for key in sorted(list(bg.meta.keys())):
                    val = bg.meta[key]
                    data.append("{}={}".format(key, val))
                return hash_obj(data)
        elif (isinstance(bg, list) and
              isinstance(bg[0], qpimage.QPImage)):
            # List of QPImage
            data = []
            for bgii in bg:
                data.append(self._compute_bgid(bgii))
            return hash_obj(data)
        elif (isinstance(bg, SeriesData) and
              (len(bg) == 1 or
               len(bg) == len(self))):
            # DataSet
            return bg.identifier
        else:
            raise ValueError("Unknown background data type: {}".format(bg))

    @functools.lru_cache(maxsize=32)
    def _identifier_data(self):
        data = []
        # data
        if isinstance(self.path, io.IOBase):
            self.path.seek(0)
            data.append(self.path.read(50 * 1024))
        else:
            with self.path.open("rb") as fd:
                data.append(fd.read(50 * 1024))
            data.append(self.path.stat().st_size)
        data += self._identifier_meta()
        return hash_obj(data)

    @functools.lru_cache(maxsize=32)
    def _identifier_meta(self):
        data = []
        # meta data
        for key in sorted(list(self.meta_data.keys())):
            value = self.meta_data[key]
            data.append("{}={}".format(key, value))
        # hologram info
        for key in sorted(list(self.holo_kw.keys())):
            value = self.holo_kw[key]
            data.append("{}={}".format(key, value))
        return hash_obj(data)

    @property
    def identifier(self):
        """Return a unique identifier for the given data set"""
        if self.background_identifier is None:
            idsum = self._identifier_data()
        else:
            idsum = hash_obj([self._identifier_data(),
                              self.background_identifier])
        return idsum

    @property
    @functools.lru_cache()
    def shape(self):
        """Return dataset shape (lenght, image0, image1).

        This should be overridden by the subclass, because by default
        the first qpimage is used for that.
        """
        qpi0 = self.get_qpimage_raw(0)
        return len(self), qpi0.shape[0], qpi0.shape[1]

    def get_identifier(self, idx):
        """Return an identifier for the data at index `idx`

        .. versionchanged:: 0.4.2
            indexing starts at 1 instead of 0
        """
        return "{}:{}".format(self.identifier, idx + 1)

    def get_name(self, idx):
        """Return name of data at index `idx`

        .. versionchanged:: 0.4.2
            indexing starts at 1 instead of 0
        """
        return "{}:{}".format(self.path, idx + 1)

    def get_time(self, idx):
        """Return time of data at index `idx`

        Returns nan if the time is not defined"""
        # raw data
        qpi = self.get_qpimage_raw(idx)
        if "time" in qpi.meta:
            thetime = qpi.meta["time"]
        else:
            thetime = np.nan
        return thetime

    def get_qpimage(self, idx):
        """Return background-corrected QPImage of data at index `idx`"""
        # raw data
        qpi = self.get_qpimage_raw(idx)
        if "identifier" not in qpi:
            msg = "`get_qpimage_raw` does not set 'identifier' " \
                  + "in class '{}'!".format(self.__class__)
            raise KeyError(msg)
        # bg data
        if self._bgdata:
            if len(self._bgdata) == 1:
                # One background for all
                bgidx = 0
            else:
                bgidx = idx

            if isinstance(self._bgdata, SeriesData):
                # `get_qpimage` does take `idx`
                bg = self._bgdata.get_qpimage_raw(bgidx)
            else:
                # `self._bgdata` is a QPImage
                bg = self._bgdata[bgidx]
            qpi.set_bg_data(bg_data=bg)
        return qpi

    @abc.abstractmethod
    def get_qpimage_raw(self, idx):
        """Return QPImage without background correction

        Note that this method must always return a QPImage instance with
        the "identifier" metadata key set!
        """

    def saveh5(self, h5file, qpi_slice=None, series_slice=None,
               time_interval=None, count=None, max_count=None):
        """Save the data set as an HDF5 file (qpimage.QPSeries format)

        Parameters
        ----------
        h5file: str, pathlib.Path, or h5py.Group
            Where to store the series data
        qpi_slice: tuple of (slice, slice)
            If not None, only store a slice of each QPImage
            in `h5file`. A value of None is equivalent to
            ``(slice(0, -1), slice(0, -1))``.
        series_slice: slice
            If None, save the entire series, otherwise only save
            the images specified by this slice.
        time_interval: tuple of (float, float)
            If not None, only stores QPImages that were recorded
            within the given time interval.
        count, max_count: multiprocessing.Value
            Can be used to monitor the progress of the algorithm.
            Initially, the value of `max_count.value` is incremented
            by the total number of steps. At each step, the value
            of `count.value` is incremented.

        Notes
        -----
        The series "identifier" meta data is only set when all
        of `qpi_slice`, `series_slice`, and `time_interval`
        are None.
        """
        # set up slice to export
        if series_slice is None:
            sl = range(len(self))
        else:
            sl = range(series_slice.start, series_slice.stop)
        # set up time interval
        if time_interval is None:
            ta = -np.inf
            tb = np.inf
        else:
            ta, tb = time_interval
        # set max_count according to slice
        if max_count is not None:
            max_count.value += len(sl)

        qpskw = {"h5file": h5file,
                 "h5mode": "w",
                 }

        if (qpi_slice is None and
            series_slice is None and
                time_interval is None):
            # Only add series identifier if series complete.
            # (We assume that if any of the above kwargs is set,
            # the series data is somehow modified)
            qpskw["identifier"] = self.identifier

        with qpimage.QPSeries(**qpskw) as qps:
            increment = 0
            for ii in sl:
                ti = self.get_time(ii)
                if ti < ta or ti > tb:
                    # Not part of the series
                    pass
                else:
                    increment += 1
                    if increment == 1 or len(self._bgdata) != 1:
                        # initial image or series data where each image
                        # has a unique background image
                        qpi = self.get_qpimage(ii)
                        if qpi_slice is not None:
                            qpi = qpi[qpi_slice]
                        qps.add_qpimage(qpi)
                    else:
                        # hard-link the background data
                        qpiraw = self.get_qpimage_raw(ii)
                        if qpi_slice is not None:
                            qpiraw = qpiraw[qpi_slice]
                        qps.add_qpimage(qpiraw, bg_from_idx=0)
                if count is not None:
                    count.value += 1

    def set_bg(self, dataset):
        """Set background data

        Parameters
        ----------
        dataset: `DataSet`, `qpimage.QPImage`, or int
            If the ``len(dataset)`` matches ``len(self)``,
            then background correction is performed
            element-wise. Otherwise, ``len(dataset)``
            must be one and is used for all data of ``self``.

        See Also
        --------
        get_qpimage: obtain the background corrected QPImage
        """
        if isinstance(dataset, qpimage.QPImage):
            # Single QPImage
            self._bgdata = [dataset]
        elif (isinstance(dataset, list) and
              len(dataset) == len(self) and
              isinstance(dataset[0], qpimage.QPImage)):
            # List of QPImage
            self._bgdata = dataset
        elif (isinstance(dataset, SeriesData) and
              (len(dataset) == 1 or
               len(dataset) == len(self))):
            # DataSet
            self._bgdata = dataset
        else:
            raise ValueError("Bad length or type for bg: {}".format(dataset))

        self.background_identifier = self._compute_bgid()

    @staticmethod
    @abc.abstractmethod
    def verify(path):
        """Verify that `path` has this file format

        Returns `True` if the file format matches.
        The implementation of this method should be fast and
        memory efficient, because e.g. the "GroupFolder" file
        format depends on it.
        """


class SingleData(SeriesData):
    """Single data file format base class

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the experimental data file.
    meta_data: dict
        Dictionary containing meta data.
        see :py:class:`qpimage.META_KEYS`.
    as_type: str
        Defines the data type that the input data is casted to.
        The default is "float32" which saves memory. If high
        numerical accuracy is required (does not apply for a
        simple 2D phase analysis), set this to double precision
        ("float64").
    """
    __meta__ = abc.ABCMeta
    is_series = False

    def __len__(self):
        return 1

    def get_identifier(self, idx=0):
        return self.identifier

    def get_name(self, idx=0):
        return super(SingleData, self).get_name(idx=0)

    def get_qpimage(self, idx=0):
        return super(SingleData, self).get_qpimage(idx=0)

    @abc.abstractmethod
    def get_qpimage_raw(self, idx=0):
        """QPImage without background correction"""

    def get_time(self, idx=0):
        """Time of the data

        Returns nan if the time is not defined
        """
        thetime = super(SingleData, self).get_time(idx=0)
        return thetime


def hash_obj(data, maxlen=5):
    hasher = hashlib.md5()
    tohash = obj2bytes(data)
    hasher.update(tohash)
    return hasher.hexdigest()[:maxlen]


def obj2bytes(data):
    tohash = []
    if isinstance(data, (tuple, list)):
        for item in data:
            tohash.append(obj2bytes(item))
    elif isinstance(data, str):
        tohash.append(data.encode("utf-8"))
    elif isinstance(data, bytes):
        tohash.append(data)
    elif isinstance(data, np.ndarray):
        tohash.append(data.tobytes())
    elif isinstance(data, int):
        tohash.append(bytes(data))
    else:
        msg = "No rule to convert to bytes: {}".format(data)
        raise NotImplementedError(msg)
    return b"".join(tohash)



import functools
import warnings

import h5py
import qpimage

from ..dataset import SeriesData


class HyperSpyNoDataFoundError(BaseException):
    pass


class WrongSignalTypeWarnging(UserWarning):
    pass


class SeriesHdf5HyperSpy(SeriesData):
    """HyperSpy hologram series (HDF5 format)

    HyperSpy has its :ref:`own implementation
    <hyperspy:electron-holography-label>` to read this file format.
    """
    storage_type = "raw-oah"
    priority = -9  # higher priority, because it's fast

    def __len__(self):
        return len(self._get_experiments())

    def _check_experiment(self, name):
        """Check the signal type "hologram" of the experiment

        Returns
        -------
        True, if the signal type is supported, False otherwise

        Raises
        ------
        Warning if the signal type is not supported
        """
        with h5py.File(name=self.path, mode="r") as h5:
            sigpath = "/Experiments/{}/metadata/Signal".format(name)
            signal_type = h5[sigpath].attrs["signal_type"]
        if signal_type != "hologram":
            msg = "Signal type '{}' not supported: {}[{}]".format(signal_type,
                                                                  self.path,
                                                                  name)
            warnings.warn(msg, WrongSignalTypeWarnging)
        return signal_type == "hologram"

    @functools.lru_cache(maxsize=5)
    def _get_experiments(self):
        """Get all experiments from the hdf5 file"""
        explist = []
        with h5py.File(name=self.path, mode="r") as h5:
            if "Experiments" not in h5:
                msg = "Group 'Experiments' not found in {}.".format(self.path)
                raise HyperSpyNoDataFoundError(msg)
            for name in h5["Experiments"]:
                # check experiment
                if self._check_experiment(name):
                    explist.append(name)
        explist.sort()
        if not explist:
            # if this error is raised, the signal_type is probably not
            # set to "hologram".
            msg = "No supported data found: {}".format(self.path)
            raise HyperSpyNoDataFoundError(msg)
        return explist

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        name = self._get_experiments()[idx]
        with h5py.File(name=self.path, mode="r") as h5:
            exp = h5["Experiments"][name]
            # hologram data
            data = exp["data"][:]
            # resolution
            rx = exp["axis-0"].attrs["scale"]
            ry = exp["axis-1"].attrs["scale"]
            if rx != ry:
                raise NotImplementedError("Only square pixels supported!")
            runit = exp["axis-0"].attrs["units"]
            if runit == "nm":
                pixel_size = rx * 1e-9
            else:
                raise NotImplementedError("Units '{}' not implemented!")

        meta_data = {"pixel size": pixel_size}
        meta_data.update(self.meta_data)

        qpi = qpimage.QPImage(data=data,
                              which_data="raw-oah",
                              meta_data=meta_data,
                              holo_kw=self.holo_kw,
                              h5dtype=self.as_type)
        # set identifier
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` has the HyperSpy file format"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
        except (OSError, IsADirectoryError):
            pass
        else:
            if ("file_format" in h5.attrs and
                h5.attrs["file_format"].lower() == "hyperspy" and
                    "Experiments" in h5):
                valid = True
            h5.close()
        return valid



import pathlib
import tempfile

import numpy as np

import qpformat


def test_load_phase():
    # generate test data
    tf = pathlib.Path(tempfile.mktemp(suffix=".npy", prefix="qpformat_test_"))
    phase = np.ones((20, 20), dtype=float)
    phase *= np.linspace(0, .3, 20).reshape(-1, 1)
    np.save(tf, phase)

    ds = qpformat.load_data(tf, as_type="float64")
    assert np.allclose(ds.get_qpimage().pha, phase, atol=1e-15, rtol=0)
    assert ds.path.samefile(tf)
    assert "SingleNpyNumpy" in ds.__repr__()


def test_load_field():
    # generate test data
    tf = pathlib.Path(tempfile.mktemp(suffix=".npy", prefix="qpformat_test_"))
    phase = np.ones((20, 20), dtype=float)
    phase *= np.linspace(0, .3, 20).reshape(-1, 1)
    amplitude = np.ones((20, 20), dtype=float)
    amplitude += np.linspace(-.1, .12, 20).reshape(1, -1)
    field = amplitude * np.exp(1j * phase)
    np.save(tf, field)

    ds = qpformat.load_data(tf, as_type="float64")
    assert np.allclose(ds.get_qpimage().pha, phase, atol=1e-15, rtol=0)
    assert np.allclose(ds.get_qpimage().amp, amplitude, atol=1e-15, rtol=0)


def test_returned_identifier():
    # generate test data
    tf = pathlib.Path(tempfile.mktemp(suffix=".npy", prefix="qpformat_test_"))
    phase = np.ones((20, 20), dtype=float)
    phase *= np.linspace(0, .3, 20).reshape(-1, 1)
    np.save(tf, phase)

    ds = qpformat.load_data(tf, as_type="float64")
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from functools import lru_cache
from os.path import commonprefix
import pathlib

from .dataset import SeriesData, hash_obj
from .dataset import SingleData  # noqa:F401 (user convenience)
from . import fmts_ready, fmts_raw_oah


class BadFileFormatError(BaseException):
    """Base class for file format errors"""


class MultipleFormatsNotSupportedError(BadFileFormatError):
    """Used when a folder contains series file formats

    (see `GitHub issue #1 <https://github.com/RI-imaging/qpformat/issues/1>`__)
    """
    pass


class UnknownFileFormatError(BadFileFormatError):
    """Used when a file format could not be detected"""
    pass


class WrongFileFormatError(BadFileFormatError):
    """Used when a wrong file format is used"""
    pass


class SeriesFolder(SeriesData):
    """Folder-based wrapper file format"""
    # storage_type is implemented as a property
    priority = -3  # higher than zip file format (issues on Windows)

    def __init__(self, *args, **kwargs):
        super(SeriesFolder, self).__init__(*args, **kwargs)
        self._files = None
        self._formats = None
        self._dataset = None

    def __len__(self):
        return len(self.files)

    @lru_cache(maxsize=32)
    def _get_cropped_file_names(self):
        """self.files with common path prefix/suffix removed"""
        files = [ff.name for ff in self.files]
        prefix = commonprefix(files)
        suffix = commonprefix([f[::-1] for f in files])[::-1]
        cropped = [f[len(prefix):-len(suffix)] for f in files]
        return cropped

    def _get_dataset(self, idx):
        if self._dataset is None:
            self._dataset = [None] * len(self)
        if self._dataset[idx] is None:
            format_class = formats_dict[self._formats[idx]]
            self._dataset[idx] = format_class(path=self._files[idx],
                                              meta_data=self.meta_data,
                                              as_type=self.as_type,
                                              holo_kw=self.holo_kw)
        if len(self._dataset[idx]) != 1:
            msg = "Multiple images per file are not supported in the " \
                  + "SeriesFolder file format! Besides the fact that it " \
                  + "would add unnecessary complexity, it is also really a " \
                  + "bad idea to do this. Please restructure your " \
                  + "workflow accordingly. The offending file is " \
                  + "'{}'.".format(self.files[idx])
            raise NotImplementedError(msg)
        return self._dataset[idx]

    @lru_cache(maxsize=32)
    def _identifier_data(self):
        """Return a unique identifier for the folder data"""
        # Use only file names
        data = [ff.name for ff in self.files]
        data.sort()
        # also use the folder name
        data.append(self.path.name)
        # add meta data
        data += self._identifier_meta()
        return hash_obj(data)

    @staticmethod
    @lru_cache(maxsize=32)
    def _search_files(path):
        """Search a folder for data files

        .. versionchanged:: 0.6.0
            `path` is not searched recursively anymore
        """
        path = pathlib.Path(path)
        fifo = []

        for fp in path.glob("*"):
            if fp.is_dir():
                continue
            for fmt in formats:
                # series data is not supported in SeriesFolder
                if not fmt.is_series and fmt.verify(fp):
                    fifo.append((fp, fmt.__name__))
                    break

        # ignore qpimage formats if multiple formats were
        # detected.
        theformats = [ff[1] for ff in fifo]
        formset = set(theformats)
        if len(formset) > 1:
            fmts_qpimage = ["SingleHdf5Qpimage", "SeriesHdf5Qpimage"]
            fifo = [ff for ff in fifo if ff[1] not in fmts_qpimage]
        # ignore raw tif files if single_tif_phasics is detected
        if len(formset) > 1 and "SingleTifPhasics" in theformats:
            fmts_badtif = "SingleTifHolo"
            fifo = [ff for ff in fifo if ff[1] not in fmts_badtif]
        # otherwise, prevent multiple file formats
        theformats2 = [ff[1] for ff in fifo]
        formset2 = set(theformats2)
        if len(formset2) > 1:
            msg = "Qpformat does not support multiple different file " \
                  + "formats within one directory: {}".format(formset2)
            raise MultipleFormatsNotSupportedError(msg)
        # sort the lists
        fifo = sorted(fifo)
        return fifo

    @property
    def files(self):
        """List of files (only supported file formats)"""
        if self._files is None:
            fifo = SeriesFolder._search_files(self.path)
            self._files = [ff[0] for ff in fifo]
            self._formats = [ff[1] for ff in fifo]
        return self._files

    @property
    def storage_type(self):
        """The storage type depends on the wrapped file format"""
        ds = self._get_dataset(0)
        return ds.storage_type

    def get_identifier(self, idx):
        """Return an identifier for the data at index `idx`

        .. versionchanged:: 0.4.2
            indexing starts at 1 instead of 0
        """
        name = self._get_cropped_file_names()[idx]
        return "{}:{}:{}".format(self.identifier, name, idx + 1)

    def get_name(self, idx):
        """Return name of data at index `idx`

        .. versionadded:: 0.4.2
        """
        return "{}".format(self.path / self.files[idx])

    def get_time(self, idx):
        ds = self._get_dataset(idx)
        return ds.get_time()

    def get_qpimage_raw(self, idx):
        """Return QPImage without background correction"""
        ds = self._get_dataset(idx)
        qpi = ds.get_qpimage_raw()
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify folder file format

        The folder file format is only valid when
        there is only one file format present.
        """
        valid = True
        fifo = SeriesFolder._search_files(path)
        # dataset size
        if len(fifo) == 0:
            valid = False
        # number of different file formats
        fifmts = [ff[1] for ff in fifo]
        if len(set(fifmts)) != 1:
            valid = False
        return valid


# the order is important
formats = sorted(
    [SeriesFolder]
    + fmts_ready.registered_formats
    + fmts_raw_oah.registered_formats,
    key=lambda x: x.priority)

# convenience dictionary
formats_dict = {}
for fmt in formats:
    formats_dict[fmt.__name__] = fmt



"""
The test file "series_phasics.zip" was created from an original
Phasics zip file. The "ConfigStack.txt" was adapted to reflect
only three measured images. All data files that are not relevant
for qpformat are included as empty files to reflect the original
zip file structure. The "SID PHA*.tif" files were created with
the script given in the doc string of "test_single_tif_phasics.py".
"""
import pathlib

import numpy as np

import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_load_data():
    path = datapath / "series_phasics.zip"
    ds = qpformat.load_data(path)
    assert ds.path == path.resolve()
    assert "SeriesZipTifPhasics" in ds.__repr__()


def test_data_content():
    path = datapath / "series_phasics.zip"
    ds = qpformat.load_data(path)
    assert len(ds) == 3
    assert ds.get_time(0) == 1461949418.29027
    assert ds.get_time(1) == 1461949418.62727
    assert ds.get_time(2) == 1461949419.11427
    qpi = ds.get_qpimage(0)
    assert qpi.meta["wavelength"] == 550e-9
    assert np.allclose(qpi.amp.max(), 183.96992660669972)
    assert np.allclose(qpi.pha.max() - qpi.pha.min(), 0.18902349472045898)


def test_returned_identifier():
    path = datapath / "series_phasics.zip"
    ds = qpformat.load_data(path)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import numpy as np

import qpformat


data_path = pathlib.Path(__file__).parent / "data"


def test_series_raw_oah():
    ds = qpformat.load_data(data_path / "single_hdf5_raw-oah.h5")
    assert len(ds) == 1
    assert ds.format == "SingleHDF5RawOAH"
    assert ds.meta_data["wavelength"] == 532e-9
    assert ds.meta_data["numerical aperture"] == 1.0
    assert np.allclose(ds.meta_data["pos x"],
                       -0.0002045580000000009)
    assert ds.get_qpimage().meta["identifier"] == "d94c8"



import pathlib

import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_basic():
    path = datapath / "single_holo.tif"
    ds = qpformat.load_data(path)
    # basic tests
    assert ds.storage_type == "raw-oah"
    assert not ds.is_series
    assert ds.path == path.resolve()
    assert "SingleTifHolo" in ds.__repr__()


def test_identifier():
    path = datapath / "single_holo.tif"
    ds = qpformat.load_data(path, holo_kw={"sideband": 1})
    assert ds.identifier == "b65c8"


def test_load_data():
    path = datapath / "single_holo.tif"
    ds = qpformat.load_data(path)
    # basic tests
    qpi = ds.get_qpimage(0)
    assert qpi.shape == (238, 267)


def test_returned_identifier():
    path = datapath / "single_holo.tif"
    ds = qpformat.load_data(path)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

import numpy as np
import qpformat
from qpformat.file_formats import WrongFileFormatError


data_dir = pathlib.Path(__file__).parent / "data"


def test_basic():
    ds = qpformat.load_data(data_dir / "series_hdf5_meep.h5")
    qpi = ds.get_qpimage(0)

    assert len(ds) == 18
    # - default wavelength is 500nm
    # - original sampling was 13 px / wavelength
    # - data were shrinked by taking every 8th pixel
    assert np.allclose(qpi["pixel size"], 500e-9 / 13 * 8,
                       rtol=0, atol=1e-9)
    assert qpi.meta["medium index"] == 1.333
    assert qpi.meta["wavelength"] == 500e-9
    assert qpi.meta["focus"] == 500e-9 * 111 / 13
    assert ds.shape == (18, 47, 47)
    assert qpi.meta["angle"] == 0

    qpi2 = ds.get_qpimage(11)
    assert np.allclose(
        qpi2.meta["angle"],
        2 * np.pi / len(ds) * 11,
        atol=1e-6,
        rtol=0)

    assert qpi2.meta["time"] == 11


def test_returned_identifier():
    ds = qpformat.load_data(data_dir / "series_hdf5_meep.h5")
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


def test_user_defined_wavelength():
    """Special for FDTD data, because it is actually unit-less"""
    ds = qpformat.load_data(data_dir / "series_hdf5_meep.h5",
                            meta_data={"wavelength": 550e-9})
    qpi = ds.get_qpimage(0)

    # - custom wavelength is 550nm
    # - original sampling was 13 px / wavelength
    # - data were shrinked by taking every 8th pixel
    assert np.allclose(qpi["pixel size"], 550e-9 / 13 * 8,
                       rtol=0, atol=1e-9)
    assert qpi.meta["focus"] == 550e-9 * 111 / 13


def test_wrong_format():
    path = data_dir / "single_qpimage.h5"
    try:
        qpformat.load_data(path, fmt="SeriesHDF5SinogramMeep")
    except WrongFileFormatError:
        pass
    else:
        raise ValueError("qpimage data is not meep sinogram data")



import copy
from os import fspath
import pathlib

import numpy as np
import qpimage
import tifffile

from ..dataset import SingleData


class SingleTifHolo(SingleData):
    """Off-axis hologram image (TIFF format)"""
    storage_type = "raw-oah"

    @staticmethod
    def _get_tif(path):
        if hasattr(path, "seek"):  # opened file
            # Seek open file zero to avoid error in tifffile:
            # "ValueError: invalid TIFF file"
            path.seek(0)
        else:  # path
            path = fspath(path)
        return tifffile.TiffFile(path)

    def get_time(self, idx=0):
        """Time of the TIFF file

        Currently, only the file modification time is supported.
        Note that the modification time of the TIFF file is
        dependent on the file system and may have temporal
        resolution as low as 3 seconds.
        """
        if isinstance(self.path, pathlib.Path):
            thetime = self.path.stat().st_mtime
        else:
            thetime = np.nan
        return thetime

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        # Load experimental data
        with SingleTifHolo._get_tif(self.path) as tf:
            holo = tf.pages[0].asarray()
        meta_data = copy.copy(self.meta_data)
        qpi = qpimage.QPImage(data=holo,
                              which_data="raw-oah",
                              meta_data=meta_data,
                              holo_kw=self.holo_kw,
                              h5dtype=self.as_type)
        # set identifier
        qpi["identifier"] = self.get_identifier()
        qpi["time"] = self.get_time()
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` is a valid TIFF file"""
        valid = False
        try:
            tf = SingleTifHolo._get_tif(path)
        except (ValueError, IsADirectoryError, KeyError,
                tifffile.tifffile.TiffFileError):
            pass
        else:
            if len(tf.pages) == 1:
                valid = True
            if not hasattr(path, "seek"):
                tf.close()
        return valid



import io
import functools
import zipfile

from ..dataset import SeriesData
from .single_tif_phasics import SingleTifPhasics


class SeriesZipTifPhasics(SeriesData):
    """Phasics series data (zipped "SID PHA*.tif" files)

    The data are stored as multiple TIFF files
    (:class:`qpformat.file_formats.SingleTifPhasics`) in a zip file.
    """
    storage_type = "phase,intensity"
    priority = -1  # should get higher priority than SeriesZipTifHolo

    def __init__(self, *args, **kwargs):
        super(SeriesZipTifPhasics, self).__init__(*args, **kwargs)
        self._files = None
        self._dataset = None

    def __len__(self):
        return len(self.files)

    def _get_dataset(self, idx):
        if self._dataset is None:
            self._dataset = [None] * len(self)
        if self._dataset[idx] is None:
            # Use ``zipfile.ZipFile.open`` to return an open file
            zf = zipfile.ZipFile(self.path)
            pt = zf.open(self.files[idx])
            fd = io.BytesIO(pt.read())
            self._dataset[idx] = SingleTifPhasics(path=fd,
                                                  meta_data=self.meta_data,
                                                  as_type=self.as_type)
        assert len(self._dataset[idx]) == 1, "unknown phasics tif file"
        return self._dataset[idx]

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def _index_files(path):
        """Search zip file for SID PHA files"""
        with zipfile.ZipFile(path) as zf:
            names = sorted(zf.namelist())
            names = [nn for nn in names if nn.endswith(".tif")]
            names = [nn for nn in names if nn.startswith("SID PHA")]
            phasefiles = []
            for name in names:
                with zf.open(name) as pt:
                    fd = io.BytesIO(pt.read())
                    if SingleTifPhasics.verify(fd):
                        phasefiles.append(name)
            return phasefiles

    @property
    def files(self):
        """List of Phasics tif file names in the input zip file"""
        if self._files is None:
            self._files = SeriesZipTifPhasics._index_files(self.path)
        return self._files

    def get_qpimage_raw(self, idx):
        """Return QPImage without background correction"""
        ds = self._get_dataset(idx)
        qpi = ds.get_qpimage_raw()
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    def get_time(self, idx):
        # Obtain the time from the tif meta data
        ds = self._get_dataset(idx)
        return ds.get_time()

    @staticmethod
    def verify(path):
        """Verify that `path` is a zip file with Phasics TIFF files"""
        valid = False
        try:
            zf = zipfile.ZipFile(path)
        except (zipfile.BadZipfile, IsADirectoryError):
            pass
        else:
            names = sorted(zf.namelist())
            names = [nn for nn in names if nn.endswith(".tif")]
            names = [nn for nn in names if nn.startswith("SID PHA")]
            for name in names:
                with zf.open(name) as pt:
                    fd = io.BytesIO(pt.read())
                    if SingleTifPhasics.verify(fd):
                        valid = True
                        break
            zf.close()
        return valid



from ._version import version as __version__  # noqa: F401
from .core import load_data  # noqa: F401
from . import file_formats  # noqa: F401
from .file_formats import BadFileFormatError  # noqa: F401



import copy

import h5py
import numpy as np
import qpimage

from ..dataset import SingleData


class SingleHDF5RawOAH(SingleData):
    """Raw off-axis holography data stored in an HDF5 file"""
    storage_type = "raw-oah"
    priority = -10  # higher priority, because it's fast

    def __init__(self, *args, **kwargs):
        super(SingleHDF5RawOAH, self).__init__(*args, **kwargs)
        # update meta data
        with h5py.File(self.path, mode="r") as h5:
            attrs = dict(h5["0"].attrs)
        for key in qpimage.meta.META_KEYS:
            if (key not in self.meta_data
                    and key in attrs):
                self.meta_data[key] = attrs[key]

    def get_time(self, idx=0):
        """Time for each dataset"""
        with h5py.File(self.path) as h5:
            ds = h5["0"]
            thetime = ds.attrs.get("time", np.nan)
        return thetime

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        # Load experimental data
        with h5py.File(self.path) as h5:
            holo = h5["0"][:]
        meta_data = copy.copy(self.meta_data)
        qpi = qpimage.QPImage(data=holo,
                              which_data="raw-oah",
                              meta_data=meta_data,
                              holo_kw=self.holo_kw,
                              h5dtype=self.as_type)
        # set identifier
        qpi["identifier"] = self.get_identifier()
        qpi["time"] = self.get_time()
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` is in the correct file format"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
        except (OSError,):
            pass
        else:
            if (h5.attrs.get("file_format", "") == "qpformat"
                and h5.attrs.get("imaging_modality", "") ==
                    "off-axis holography"
                    and "0" in h5
                    and "1" not in h5):
                valid = True
            h5.close()
        return valid



import h5py
import qpimage

from ..dataset import SeriesData


class SeriesHdf5Qpimage(SeriesData):
    """Qpimage series (HDF5 format)"""
    storage_type = "phase,amplitude"
    priority = -9  # higher priority, because it's fast

    def __init__(self, *args, **kwargs):
        super(SeriesHdf5Qpimage, self).__init__(*args, **kwargs)
        self._dataset = None
        self._init_meta()

    def __len__(self):
        with self._qpseries() as qps:
            return len(qps)

    def _init_meta(self):
        # update meta data
        with h5py.File(self.path, mode="r") as h5:
            attrs = dict(h5["qpi_0"].attrs)
        for key in qpimage.meta.META_KEYS:
            if (key not in self.meta_data
                    and key not in ["time"]  # do not override time
                    and key in attrs):
                self.meta_data[key] = attrs[key]

    def _qpseries(self):
        return qpimage.QPSeries(h5file=self.path, h5mode="r")

    def get_qpimage(self, idx):
        """Return background-corrected QPImage of data at index `idx`"""
        if self._bgdata:
            # The user has explicitly chosen different background data
            # using `get_qpimage_raw`.
            qpi = super(SeriesHdf5Qpimage, self).get_qpimage(idx)
        else:
            # We can use the background data stored in the qpimage hdf5 file
            with self._qpseries() as qps:
                qpi = qps.get_qpimage(index=idx).copy()
            # Force meta data
            for key in self.meta_data:
                qpi[key] = self.meta_data[key]
            # set identifier
            qpi["identifier"] = self.get_identifier(idx)
        return qpi

    def get_qpimage_raw(self, idx):
        """Return QPImage without background correction"""
        with self._qpseries() as qps:
            qpi = qps.get_qpimage(index=idx).copy()
        # Remove previously performed background correction
        qpi.set_bg_data(None)
        # Force meta data
        for key in self.meta_data:
            qpi[key] = self.meta_data[key]
        # set identifier
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` has the qpimage series file format"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
            qpi0 = h5["qpi_0"]
        except (OSError, KeyError):
            pass
        else:
            if ("qpimage version" in qpi0.attrs and
                "phase" in qpi0 and
                "amplitude" in qpi0 and
                "bg_data" in qpi0["phase"] and
                    "bg_data" in qpi0["amplitude"]):
                valid = True
        return valid


class SeriesHdf5QpimageSubjoined(SeriesHdf5Qpimage):
    """Subjoined qpimage series (HDF5 format), may contain other data"""

    def _init_meta(self):
        # update meta data
        with h5py.File(self.path, mode="r") as h5:
            attrs = dict(h5["qpseries/qpi_0"].attrs)
        for key in qpimage.meta.META_KEYS:
            if (key not in self.meta_data
                    and key in attrs):
                self.meta_data[key] = attrs[key]

    def _qpseries(self):
        h5 = h5py.File(self.path, mode="r")
        return qpimage.QPSeries(h5file=h5["qpseries"])

    @staticmethod
    def verify(path):
        """Verify that `path` has the qpimage series file format"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
            qps = h5["qpseries"]
            qpi0 = qps["qpi_0"]
        except (OSError, KeyError):
            pass
        else:
            if ("qpimage version" in qpi0.attrs and
                "phase" in qpi0 and
                "amplitude" in qpi0 and
                "bg_data" in qpi0["phase"] and
                    "bg_data" in qpi0["amplitude"]):
                valid = True
            h5.close()
        return valid



import pathlib
import tempfile

import numpy as np

import qpimage
import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_identifier():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi["identifier"] = "an extremely important string"
    qpi.copy(tf)

    ds1 = qpformat.load_data(path)
    ds2 = qpformat.load_data(tf)

    assert ds1.identifier != ds2.identifier
    # individual identifiers are not extracted anymore as of
    # qpformat version 0.3.4
    assert "an extremely important string" not in ds2.identifier
    assert ds1.identifier == ds1.get_identifier()
    assert ds2.identifier == ds2.get_identifier()


def test_load_data():
    path = datapath / "single_qpimage.h5"
    ds = qpformat.load_data(path)
    assert ds.path == path.resolve()
    assert np.isnan(ds.get_time()), "original file has no time metadata"
    assert "SingleHdf5Qpimage" in ds.__repr__()
    qpd = ds.get_qpimage()
    qpi = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi["identifier"] = "test"
    assert qpd["identifier"] != qpi["identifier"]
    assert qpd == qpi
    assert qpd.shape == qpi.shape
    assert np.allclose(qpd.amp, qpi.amp)
    assert np.allclose(qpd.pha, qpi.pha)


def test_meta_extraction():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi["wavelength"] = 345e-9
    qpi["pixel size"] = .1e-6
    qpi["medium index"] = 1.336
    qpi.copy(tf)

    ds = qpformat.load_data(tf)

    assert ds.meta_data["wavelength"] == 345e-9
    assert ds.meta_data["pixel size"] == .1e-6
    assert ds.meta_data["medium index"] == 1.336


def test_meta_override():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi.copy(tf)

    wl = 333e-9
    px = .111
    ds = qpformat.load_data(tf, meta_data={"wavelength": wl,
                                           "pixel size": px})

    assert ds.meta_data["wavelength"] == wl
    assert ds.meta_data["pixel size"] == px

    qpi_ds = ds.get_qpimage(0)
    assert qpi_ds["wavelength"] == wl
    assert qpi_ds["pixel size"] == px


def test_returned_identifier():
    path = datapath / "single_qpimage.h5"
    ds = qpformat.load_data(path)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from .series_hdf5_meep import SeriesHDF5SinogramMeep
from .series_hdf5_qpimage import SeriesHdf5Qpimage, SeriesHdf5QpimageSubjoined
from .series_zip_tif_phasics import SeriesZipTifPhasics
from .single_hdf5_qpimage import SingleHdf5Qpimage
from .single_npy_numpy import SingleNpyNumpy
from .single_tif_phasics import SingleTifPhasics


registered_formats = [
    SeriesHDF5SinogramMeep,
    SeriesHdf5Qpimage,
    SeriesHdf5QpimageSubjoined,
    SeriesZipTifPhasics,
    SingleHdf5Qpimage,
    SingleNpyNumpy,
    SingleTifPhasics,
]



"""
This test uses a cropped version of an original phasics tif file
(data/single_phasics.tif), which was created using this script
(the tags are important):

    import tifffile

    a = tifffile.TiffFile("in.tif")

    x1 = 100
    x2 = 180
    y1 = 140
    y2 = 260

    for ii in range(3):
        data = a.pages[ii].asarray()[x1:x2, y1:y2]

        extratags = []
        for tag in ["61237",
                    "61238",
                    "61239",
                    "61240",
                    "61241",
                    "61242",
                    "61243",
                    "max_sample_value",
                    "min_sample_value",
                    ]:
            tag = a.pages[ii].tags[tag]
            extratags.append((tag.code,
                              tag.dtype.strip("1234567890"),
                              1,
                              tag.value,
                              False))

        tifffile.imsave(file="single_phasics.tif",
                        data=data,
                        extratags=extratags,
                        append=True)
"""
import pathlib

import numpy as np

import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_load_data():
    path = datapath / "single_phasics.tif"
    ds = qpformat.load_data(path)
    assert pathlib.Path(ds.path) == path.resolve()
    assert "SingleTifPhasics" in ds.__repr__()


def test_data_content():
    path = datapath / "single_phasics.tif"
    ds = qpformat.load_data(path)
    assert ds.get_time() == 1461951095.00827
    qpi = ds.get_qpimage()
    assert qpi.meta["wavelength"] == 550e-9
    assert np.allclose(qpi.amp.max(), 188.57930365845519)
    assert np.allclose(qpi.pha.max() - qpi.pha.min(), 4.168394088745117)


def test_returned_identifier():
    path = datapath / "single_phasics.tif"
    ds = qpformat.load_data(path)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
import pathlib
import tempfile
import time
import zipfile

import qpformat


datapath = pathlib.Path(__file__).parent / "data"


DATATIME = 1529500484
# We need this due to a peculiarity of the zip file format
UTCOFFSET = time.mktime(time.gmtime(0))


def setup_test_zip(num=5):
    _fd, name = tempfile.mkstemp(suffix=".zip", prefix="qpformat_test_holo_")
    tifname = datapath / "single_holo.tif"
    # We can safely change the time, because git does not synchronize times.
    os.utime(tifname, times=(DATATIME, DATATIME))
    with zipfile.ZipFile(name, mode="w") as arc:
        for ii in range(num):
            arc.write(tifname, arcname="test_{:04d}.tif".format(ii))
    return pathlib.Path(name)


def test_basic():
    num = 5
    path = setup_test_zip(num)
    ds = qpformat.load_data(path)
    # basic tests
    assert ds.storage_type == "raw-oah"
    assert len(ds) == num
    assert ds.is_series
    assert ds.path.samefile(path)
    assert "SeriesZipTifHolo" in ds.__repr__()


def test_time():
    path = setup_test_zip(2)
    ds = qpformat.load_data(path)
    ds.get_time(0)
    # this is the creation date of "single_holo.tif"
    assert ds.get_time(0) == DATATIME - UTCOFFSET


def test_load_data():
    num = 5
    path = setup_test_zip(num)
    ds = qpformat.load_data(path)
    # basic tests
    qpi = ds.get_qpimage(0)
    assert qpi.shape == (238, 267)


def test_returned_identifier():
    path = setup_test_zip()
    ds = qpformat.load_data(path)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""command line interface"""
import argparse
import pathlib

from .core import load_data
from .file_formats import UnknownFileFormatError


def qpinfo():
    """Print information of a quantitative phase imaging dataset"""
    parser = qpinfo_parser()
    args = parser.parse_args()

    path = pathlib.Path(args.path).resolve()
    try:
        ds = load_data(path)
    except UnknownFileFormatError:
        print("Unknown file format: {}".format(path))
        return

    print("{} ({})".format(ds.__class__.__doc__, ds.__class__.__name__))
    print("- number of images: {}".format(len(ds)))
    for key in ds.meta_data:
        print("- {}: {}".format(key, ds.meta_data[key]))


def qpinfo_parser():
    descr = "Show information about a quantitative phase imaging dataset"
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('path', metavar='path', type=str,
                        help='Data path')
    return parser



"""Conversion of external holograms to .tif files

Qpformat can load hologram data from .tif image files. If your
experimental hologram data are stored in a different file format,
you can either request its implementation in qpformat by
`creating an issue <https://github.com/RI-imaging/qpformat/issues/new>`__
or you can modify this example script to your needs.

This example must be executed with a directory as an
command line argument, i.e.
``python convert_txt2tif.py /path/to/folder/``
"""
import pathlib
import sys

import numpy as np
from skimage.external import tifffile

# File names ending with these strings are ignored
# (these are files related to previous analyses)
ignore_endswith = ['.bmp', '.npy', '.opj', '.png', '.pptx', '.py', '.svg',
                   '.tif', '.txt', '_RIdist', '_parameter', '_parameter_2',
                   '_parameter_3', '_parameter_4', '_parameterdrymass',
                   '_parameter_old', '_phase', 'n_array', 'n_array_1',
                   'n_array_drymass1', 'n_array_drymass2', 'n_array_real',
                   '~', '.dat']
# uncomment this line to keep background hologram files
ignore_endswith += ['_bg']


def get_paths(folder, ignore_endswith=ignore_endswith):
    '''Return hologram file paths

    Parameters
    ----------
    folder: str or pathlib.Path
        Path to search folder
    ignore_endswith: list
        List of filename ending strings indicating which
        files should be ignored.
    '''
    folder = pathlib.Path(folder).resolve()
    files = folder.rglob("*")
    for ie in ignore_endswith:
        files = [ff for ff in files if not ff.name.endswith(ie)]
    return sorted(files)


if __name__ == "__main__":
    path = pathlib.Path(sys.argv[-1])
    if not path.is_dir():
        raise ValueError("Command line argument must be directory!")
    # output directory
    pout = path.parent / (path.name + "_tif")
    pout.mkdir(exist_ok=True)
    # get input hologram files
    files = get_paths(path)
    # conversion
    for ff in files:
        # convert image data to uint8 (most image sensors)
        hol = np.loadtxt(str(ff), dtype=np.uint8)
        tifout = str(pout / (ff.name + ".tif"))
        # compress image data
        tifffile.imsave(tifout, hol, compress=9)



"""Hologram from tif file

This example illustrates how to retrieve phase and amplitude from a
hologram stored as a tif file. The experimental hologram is a
U.S. Air Force test target downloaded from the
`Submersible Holographic Astrobiology Microscope with
Ultraresolution (SHAMU) <https://github.com/bmorris3/shampoo>`_
project :cite:`Bedrossian2017`. The values for pixel resolution,
wavelength, and reconstruction distance are taken from the
corresponding `Python example <http://shampoo.readthedocs.io/en/latest
/getting_started.html#simple-numerical-reconstruction>`_.

The object returned by the
:func:`get_qpimage <qpformat.file_formats.dataset.SingleData.get_qpimage`
function is an instance of :class:`qpimage.QPImage <qpimage.core.QPImage>`
which allows for field refocusing. The refocused QPImage is
background-corrected using a polynomial fit to the phase data at locations
where the amplitude data is not attenuated (bright regions in the amplitude
image).
"""

import urllib.request
import os

import matplotlib.pylab as plt
import qpformat


# load the experimental data
dl_loc = "https://github.com/bmorris3/shampoo/raw/master/data/"
dl_name = "USAF_test.tif"
if not os.path.exists(dl_name):
    print("Downloading {} ...".format(dl_name))
    urllib.request.urlretrieve(dl_loc + dl_name, dl_name)


ds = qpformat.load_data(dl_name,
                        # manually set meta data
                        meta_data={"pixel size": 3.45e-6,
                                   "wavelength": 405e-9,
                                   "medium index": 1},
                        # set filter size to 1/2 (defaults to 1/3)
                        # which increases the image resolution
                        holo_kw={"filter_size": .5})

# retrieve the qpimage.QPImage instance
qpi = ds.get_qpimage()
# refocus `qpi` to 0.03685m
qpi_foc = qpi.refocus(0.03685)
# perform an offset-based amplitude correction
qpi_foc.compute_bg(which_data="amplitude",
                   fit_profile="offset",
                   fit_offset="mode",
                   border_px=10,
                   )
# perform a phase correction using
# - those pixels that are not dark in the amplitude image (amp_bin) and
# - a 2D second order polynomial fit to the phase data
amp_bin = qpi_foc.amp > 1  # bright regions
qpi_foc.compute_bg(which_data="phase",
                   fit_profile="poly2o",
                   from_mask=amp_bin,
                   )

# plot results
plt.figure(figsize=(8, 3.5))
# ampltitude
ax1 = plt.subplot(121, title="amplitude")
map1 = plt.imshow(qpi_foc.amp, cmap="gray")
plt.colorbar(map1, ax=ax1, fraction=.0455, pad=0.04)
# phase in interval [-1rad, 1rad]
ax2 = plt.subplot(122, title="phase [rad]")
map2 = plt.imshow(qpi_foc.pha, vmin=-1, vmax=1)
plt.colorbar(map2, ax=ax2, fraction=.0455, pad=0.04)
# disable axes
[ax.axis("off") for ax in [ax1, ax2]]
plt.tight_layout()
plt.show()



import h5py
import qpimage

from ..dataset import SingleData


class SingleHdf5Qpimage(SingleData):
    """Qpimage single (HDF5 format)

    See the documentation of :ref:`qpimage <qpimage:index>` for more
    information.
    """
    storage_type = "phase,amplitude"
    priority = -9  # higher priority, because it's fast

    def __init__(self, *args, **kwargs):
        super(SingleHdf5Qpimage, self).__init__(*args, **kwargs)
        # update meta data
        with h5py.File(self.path, mode="r") as h5:
            attrs = dict(h5.attrs)
        for key in qpimage.meta.META_KEYS:
            if (key not in self.meta_data
                    and key in attrs):
                self.meta_data[key] = attrs[key]

    def get_qpimage(self, idx=0):
        """Return background-corrected QPImage"""
        if self._bgdata:
            # The user has explicitly chosen different background data
            # using `get_qpimage_raw`.
            qpi = super(SingleHdf5Qpimage, self).get_qpimage()
        else:
            # We can use the background data stored in the qpimage hdf5 file
            qpi = qpimage.QPImage(h5file=self.path,
                                  h5mode="r",
                                  h5dtype=self.as_type,
                                  ).copy()
            # Force meta data
            for key in self.meta_data:
                qpi[key] = self.meta_data[key]
            # set identifier
            qpi["identifier"] = self.get_identifier(idx)
        return qpi

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        qpi = qpimage.QPImage(h5file=self.path,
                              h5mode="r",
                              h5dtype=self.as_type,
                              ).copy()
        # Remove previously performed background correction
        qpi.set_bg_data(None)
        # Force meta data
        for key in self.meta_data:
            qpi[key] = self.meta_data[key]
        # set identifier
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` has the qpimage file format"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
        except (OSError,):
            pass
        else:
            if ("qpimage version" in h5.attrs and
                "phase" in h5 and
                "amplitude" in h5 and
                "bg_data" in h5["phase"] and
                    "bg_data" in h5["amplitude"]):
                valid = True
            h5.close()
        return valid



import calendar
import copy
from os import fspath
import time
import xml.etree.ElementTree as ET

import numpy as np
import qpimage
import tifffile

from ..dataset import SingleData


# baseline clamp intensity normalization for phasics tif files
INTENSITY_BASELINE_CLAMP = 150

# https://www.loc.gov/preservation/digital/formats/content/tiff_tags.shtml
TIFF_TAGS = {
    "MaxSampleValue": 281,
    }


class LoadTifPhasicsError(BaseException):
    pass


class SingleTifPhasics(SingleData):
    """Phasics image ("SID PHA*.tif")

    Notes
    -----
    - Only the processed phase data files are supported, i.e. TIFF
      file names starting with "SID PHA" exported by the commercial
      Phasics software.

    - If the "wavelength" key in `meta_data` is not set (units: [m]),
      then the wavelength is extracted from the xml data stored in
      tag "61238" of the tif file.
    """
    storage_type = "phase,intensity"

    def __init__(self, path, meta_data={}, *args, **kwargs):
        if "wavelength" not in meta_data:
            # get wavelength if not given
            wl = self._get_wavelength(path)
            if not np.isnan(wl):
                meta_data["wavelength"] = wl
            else:
                # We need the wavelength to convert OPD to phase
                msg = "'wavelength' must be specified in `meta_data`!"
                raise LoadTifPhasicsError(msg)

        super(SingleTifPhasics, self).__init__(path=path,
                                               meta_data=meta_data,
                                               *args, **kwargs)

    @staticmethod
    def _get_meta_data(path, section, name):
        with SingleTifPhasics._get_tif(path) as tf:
            meta = tf.pages[0].tags[61238].value
        meta = meta.strip("'b")
        meta = meta.replace("\\n", "\n")
        meta = meta.replace("\\r", "")
        root = ET.fromstring("<root>\n" + meta + "</root>")
        for phadata in root:
            for cluster in phadata:
                sec = cluster[0].text
                for child in cluster:
                    if len(child) == 2:
                        nm, val = child
                        # print(sec, nm.text, val.text)
                        if (sec.lower() == section and
                                nm.text.lower() == name):
                            return val.text
        else:
            return None

    @staticmethod
    def _get_tif(path):
        if hasattr(path, "seek"):  # opened file
            # Seek open file zero to avoid error in tifffile:
            # "ValueError: invalid TIFF file"
            path.seek(0)
        else:  # path
            path = fspath(path)
        return tifffile.TiffFile(path)

    def _get_wavelength(self, path):
        for section in ["analyse data", "analyse data v1"]:
            wl_str = SingleTifPhasics._get_meta_data(path=path,
                                                     section=section,
                                                     name="lambda(nm)")
            if wl_str:
                wavelength = float(wl_str) * 1e-9
                break
        else:
            wavelength = np.nan
        return wavelength

    def get_time(self, idx=0):
        """Return the time of the tif data since the epoch

        The time is stored in the "61238" tag.
        """
        timestr = SingleTifPhasics._get_meta_data(path=self.path,
                                                  section="acquisition info",
                                                  name="date & heure")
        if timestr is not None:
            timestr = timestr.replace(",", ".")
            timestrf, timeus = timestr.split(".")
            # '2016-04-29_17h31m35s.00827'
            structtime = time.strptime(timestrf,
                                       "%Y-%m-%d_%Hh%Mm%Ss")
            fracsec = float(timeus) * 1e-5
            # use calendar, because we need UTC
            thetime = calendar.timegm(structtime) + fracsec
        else:
            thetime = np.nan
        return thetime

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        # Load experimental data
        with SingleTifPhasics._get_tif(self.path) as tf:
            # page 0 contains intensity
            # page 1 contains phase in nm
            # page 2 contains phase in wavelengths
            # Intensity:
            inttags = tf.pages[0].tags
            imin = inttags["61243"].value
            imax = inttags["61242"].value
            isamp = inttags[TIFF_TAGS["MaxSampleValue"]].value
            blc = INTENSITY_BASELINE_CLAMP
            inten = tf.pages[0].asarray() * (imax - imin) / isamp + imin - blc
            inten[inten < 0] = 0
            # Phase
            # The SID4Bio records two phase images, one in wavelengths and
            # one in nanometers. Surprisingly, these two phase images are
            # not derived from the same recording. The first image (pages[1])
            # (in wavelengths) matches the intensity image (pages[0]). The
            # second image (pages[2]) is recorded at a different time point.
            # Initially, I thought it would be best to compute the phase
            # directly from the measured value in nanometers (pages[2]) using
            # the known wavelength given by the qpformat user. However, since
            # phase and amplitude won't match up in that case, the wavelength
            # phase image (pages[1]) has to be used. Since phasics uses its own
            # wavelength (set by the user in the acquisition/extraction
            # software) which might be wrong, I decided to first compute
            # the phase in nanometers from tf.pages[1] using the phasics
            # wavelength and then proceed as before, computing the phase
            # in radians using the correct, user-given wavelength.
            wl_phasics = self._get_wavelength(self.path)
            if not np.isnan(wl_phasics):
                # proceed with phase in wavelengths
                phaid = 1
            else:
                # proceed with phase in nanometers
                phaid = 2
            phatags = tf.pages[phaid].tags
            pmin = phatags["61243"].value
            pmax = phatags["61242"].value
            psamp = phatags[TIFF_TAGS["MaxSampleValue"]].value
            if psamp == 0 or pmin == pmax:
                # no phase data
                pha = np.zeros_like(inten)
            else:
                # optical path difference
                opd = tf.pages[phaid].asarray() * (pmax - pmin) / psamp + pmin
                if phaid == 1:  # convert [wavelengths] to [nm]
                    assert not np.isnan(wl_phasics)
                    opd *= wl_phasics * 1e9
                # convert from [nm] to [rad]
                pha = opd / (self.meta_data["wavelength"] * 1e9) * 2 * np.pi

        meta_data = copy.copy(self.meta_data)
        if "time" not in meta_data:
            meta_data["time"] = self.get_time()
        qpi = qpimage.QPImage(data=(pha, inten),
                              which_data="phase,intensity",
                              meta_data=meta_data,
                              h5dtype=self.as_type)
        # set identifier
        qpi["identifier"] = self.get_identifier()
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` is a phasics phase/intensity TIFF file"""
        valid = False
        try:
            tf = SingleTifPhasics._get_tif(path)
        except (ValueError, IsADirectoryError, KeyError,
                tifffile.tifffile.TiffFileError):
            pass
        else:
            if (len(tf.pages) == 3 and
                61243 in tf.pages[0].tags and
                61242 in tf.pages[0].tags and
                61238 in tf.pages[0].tags and
                61243 in tf.pages[1].tags and
                61242 in tf.pages[1].tags and
                TIFF_TAGS["MaxSampleValue"] in tf.pages[0].tags and
                (tf.pages[0].tags[61242].value !=
                 tf.pages[1].tags[61242].value)):
                valid = True
            tf.close()
        return valid



import pathlib
import tempfile
import shutil

import qpformat

data_path = pathlib.Path(__file__).parent / "data"


def test_good_folder():
    dpath = pathlib.Path(tempfile.mkdtemp(prefix="qpformat_test_"))
    shutil.copy(data_path / "single_qpimage.h5", dpath / "1.h5")
    shutil.copy(data_path / "single_qpimage.h5", dpath / "2.h5")
    ds = qpformat.load_data(dpath)
    assert ds.storage_type == "phase,amplitude"


def test_phasics():
    path = data_path / "single_phasics.tif"
    ds = qpformat.load_data(path)
    assert ds.storage_type == "phase,intensity"


def test_phasics_zip():
    path = data_path / "series_phasics.zip"
    ds = qpformat.load_data(path)
    assert ds.storage_type == "phase,intensity"


def test_qpimage():
    path = data_path / "single_qpimage.h5"
    ds = qpformat.load_data(path)
    assert ds.storage_type == "phase,amplitude"



import io
import functools
import time
import zipfile

import numpy as np

from ..dataset import SeriesData
from .single_tif_holo import SingleTifHolo


class SeriesZipTifHolo(SeriesData):
    """Off-axis hologram series (zipped TIFF files)

    The data are stored as multiple TIFF files
    (:class:`qpformat.file_formats.SingleTifHolo`) in a zip file.
    """
    storage_type = "raw-oah"

    def __init__(self, *args, **kwargs):
        super(SeriesZipTifHolo, self).__init__(*args, **kwargs)
        self._files = None
        self._dataset = None

    def __len__(self):
        return len(self.files)

    def _get_dataset(self, idx):
        if self._dataset is None:
            self._dataset = [None] * len(self)
        if self._dataset[idx] is None:
            # Use ``zipfile.ZipFile.open`` to return an open file
            zf = zipfile.ZipFile(self.path)
            pt = zf.open(self.files[idx])
            fd = io.BytesIO(pt.read())
            self._dataset[idx] = SingleTifHolo(path=fd,
                                               meta_data=self.meta_data,
                                               as_type=self.as_type,
                                               holo_kw=self.holo_kw)
        return self._dataset[idx]

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def _index_files(path):
        """Search zip file for tif files"""
        with zipfile.ZipFile(path) as zf:
            names = sorted(zf.namelist())
            names = [nn for nn in names if nn.endswith(".tif")]
            phasefiles = []
            for name in names:
                with zf.open(name) as pt:
                    fd = io.BytesIO(pt.read())
                    if SingleTifHolo.verify(fd):
                        phasefiles.append(name)
            return phasefiles

    @property
    def files(self):
        """List of hologram data file names in the input zip file"""
        if self._files is None:
            self._files = SeriesZipTifHolo._index_files(self.path)
        return self._files

    def get_time(self, idx):
        """Time for each TIFF file

        If there are no metadata keyword arguments defined for the
        TIFF file format, then the zip file `date_time` value is
        used.
        """
        # first try to get the time from the TIFF file
        # (possible meta data keywords)
        ds = self._get_dataset(idx)
        thetime = ds.get_time()
        if np.isnan(thetime):
            # use zipfile date_time
            zf = zipfile.ZipFile(self.path)
            info = zf.getinfo(self.files[idx])
            timetuple = tuple(list(info.date_time) + [0, 0, 0])
            thetime = time.mktime(timetuple)
        return thetime

    def get_qpimage_raw(self, idx):
        """Return QPImage without background correction"""
        ds = self._get_dataset(idx)
        qpi = ds.get_qpimage_raw()
        # set identifier
        qpi["identifier"] = self.get_identifier(idx)
        qpi["time"] = self.get_time(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` is a zip file containing TIFF files"""
        valid = False
        try:
            zf = zipfile.ZipFile(path)
        except (zipfile.BadZipfile, IsADirectoryError):
            pass
        else:
            names = sorted(zf.namelist())
            names = [nn for nn in names if nn.endswith(".tif")]
            for name in names:
                with zf.open(name) as pt:
                    fd = io.BytesIO(pt.read())
                    if SingleTifHolo.verify(fd):
                        valid = True
                        break
            zf.close()
        return valid



import copy
from functools import lru_cache
import pathlib

import numpy as np
import qpimage

from ..dataset import SingleData


class SingleNpyNumpy(SingleData):
    """Numpy complex field or phase data (numpy binary format)

    The experimental data given in `path` consist of a single
    2D ndarray (no pickled objects). The ndarray is either
    complex-valued (scattered field) or real-valued (phase).
    """
    # storage type is implemented as a property

    @property
    @lru_cache(maxsize=32)
    def storage_type(self):
        """Depending on input data type, the storage type is either
        "field" (complex) or "phase" (real)."""
        nf = np.load(str(self.path), mmap_mode="c", allow_pickle=False)
        if np.iscomplexobj(nf):
            st = "field"
        else:
            st = "phase"
        return st

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        # Load experimental data
        nf = np.load(str(self.path), mmap_mode="c", allow_pickle=False)
        meta_data = copy.copy(self.meta_data)
        qpi = qpimage.QPImage(data=nf,
                              which_data=self.storage_type,
                              meta_data=meta_data,
                              h5dtype=self.as_type)
        # get identifier
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify that `path` has a supported numpy file format"""
        path = pathlib.Path(path)
        valid = False
        if path.suffix == ".npy":
            try:
                nf = np.load(str(path), mmap_mode="r", allow_pickle=False)
            except (OSError, ValueError, IsADirectoryError):
                pass
            else:
                if len(nf.shape) == 2:
                    valid = True
        return valid



"""visualization of dclab definitions

Usage
-----
Directives:

Run autodoc on all supported formats:

   .. autodoc_qpformats::

Create table of all supported formats

   .. qpformats::
"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

from qpformat.file_formats import formats_dict, SeriesData


class Base(Directive):
    required_arguments = 0
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class AutodocFormats(Base):
    def generate_rst(self):
        rst = []

        refdir = [a for a in dir(SeriesData) if not a.startswith("_")]

        for key in sorted(formats_dict.keys()):
            # get any public methods that are not in the base class
            cdir = [a for a in dir(formats_dict[key]) if not a.startswith("_")]
            cdir = [a for a in cdir if a not in refdir]
            # get important public attributes
            udcand = ["is_series", "storage_type"]
            udir = [a for a in udcand if hasattr(formats_dict[key], a)]
            # remove attributes from cdir to avoid redundant docs
            [cdir.remove(a) for a in udir if a in cdir]
            rst.append("")
            # headint
            rst.append(key)
            rst.append("-"*len(key))
            # autodoc
            rst.append(".. autoclass:: qpformat.file_formats.{}".format(key))
            if cdir:
                rst.append("    :members: {}".format(", ".join(cdir)))
            if udir:
                rst.append("")
                for u in udir:
                    rst.append("    .. autoattribute:: {}".format(u))
            rst.append("")

        return rst


class Formats(Base):
    def generate_rst(self):
        rst = []

        rst.append(".. csv-table::")
        rst.append("    :header: Class, Storage type, Description".format())
        rst.append("    :widths: 2, 2, 6")
        rst.append("    :delim: tab")
        rst.append("")

        for key in sorted(formats_dict.keys()):
            cl = formats_dict[key]
            datatype = cl.storage_type
            if not isinstance(datatype, str):
                datatype = "*multiple*"
            rst.append(
                "    :class:`{key}<qpformat.file_formats.{key}>"
                "`\t {type}\t {doc}".format(
                    key=key, type=datatype, doc=cl.__doc__.split("\n", 1)[0]))

        rst.append("")

        return rst


def setup(app):
    app.add_directive('autodoc_qpformats', AutodocFormats)
    app.add_directive('qpformats', Formats)
    return {'version': '0.1'}   # identifies the version of our extension



import tempfile

import numpy as np
import qpformat.core


def test_wrong_file_format():
    tf = tempfile.mktemp(prefix="qpformat_test_")
    with open(tf, "w") as fd:
        fd.write("test")
    try:
        qpformat.core.guess_format(tf)
    except qpformat.core.UnknownFileFormatError:
        pass
    else:
        assert False, "Unknown file format was loaded!"


def test_load_with_bg():
    data = np.ones((20, 20), dtype=float)
    data *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    f_data = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(f_data, data)

    bg_data = np.ones((20, 20), dtype=float)
    bg_data *= np.linspace(0, 1.1, 20).reshape(1, -1)
    f_bg_data = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(f_bg_data, bg_data)

    ds = qpformat.core.load_data(path=f_data,
                                 bg_data=f_bg_data,
                                 as_type="float64")
    qpi = ds.get_qpimage()
    assert np.allclose(qpi.pha, data - bg_data, atol=1e-15, rtol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import shutil
import tempfile
import zipfile

import numpy as np
import qpformat

datapath = pathlib.Path(__file__).parent / "data"


def assert_path_in_list(path, path_list):
    for ff in path_list:
        if path.samefile(ff):
            break
    else:
        assert False, "{} not in {}".format(path, path_list)


def setup_folder_single_h5(size=2, tdir=None):
    path = datapath / "single_qpimage.h5"
    if tdir is None:
        tdir = tempfile.mkdtemp(prefix="qpformat_test_")
    tdir = pathlib.Path(tdir)
    files = []
    for ss in range(size):
        tpath = tdir / "data{:04d}.h5".format(ss)
        files.append(tpath)
        shutil.copy(path, tpath)
    return tdir, files


def setup_folder_single_holo(size=2, tdir=None):
    path = datapath / "single_holo.tif"
    if tdir is None:
        tdir = tempfile.mkdtemp(prefix="qpformat_test_")
    tdir = pathlib.Path(tdir)
    files = []
    for ss in range(size):
        tpath = tdir / "data{:04d}.h5".format(ss)
        files.append(tpath)
        shutil.copy(path, tpath)
    return tdir, files


def setup_folder_single_phasics_tif(tdir=None):
    path = datapath / "series_phasics.zip"
    if tdir is None:
        tdir = tempfile.mkdtemp(prefix="qpformat_test_")

    files = []
    with zipfile.ZipFile(path) as arc:
        for fn in arc.namelist():
            arc.extract(fn, path=tdir)
            if fn.startswith("SID PHA") and fn.endswith(".tif"):
                files.append(pathlib.Path(tdir) / fn)
    return tdir, files


def test_load_data():
    path, files = setup_folder_single_h5(size=2)
    ds = qpformat.load_data(path)
    # check files in folder
    assert len(ds) == 2
    for f1 in ds.files:
        assert_path_in_list(f1, files)
    # names should be different
    assert ds.get_name(0) != ds.get_name(1)
    # data should be the same
    qpi0 = ds.get_qpimage(0)
    qpi1 = ds.get_qpimage(1)
    assert qpi0 == qpi1
    assert qpi0["identifier"] != qpi1["identifier"]
    assert qpi0 == qpi1
    assert ds.get_qpimage_raw(0) == ds.get_qpimage_raw(1)
    assert ds.get_qpimage(0).shape == (50, 50)
    assert np.isnan(ds.get_time(0)), "no time defined in original file"
    # format should be right
    assert ds.verify(ds.path)
    assert ds.__class__.__name__ == "SeriesFolder"


def test_multiple_formats_phasics_tif():
    """
    Folders with phasics tif files sometimes contain raw tif files.
    This test makes sure the raw files are ignored
    """
    path, files = setup_folder_single_phasics_tif()
    ds = qpformat.load_data(path)
    for ff in ds.files:
        assert_path_in_list(ff, files)
    assert ds.verify(ds.path)
    assert ds.__class__.__name__ == "SeriesFolder"


def test_multiple_formats_phasics_tif_ignore_h5():
    """
    In folders that contain h5 files and another formats, the
    h5 files are ignored.
    """
    path, files1 = setup_folder_single_phasics_tif()
    path, _files2 = setup_folder_single_h5(tdir=path)
    ds = qpformat.load_data(path)
    for ff in ds.files:
        assert_path_in_list(ff, files1)


def test_multiple_formats_error():
    """
    Folders containing two different formats that are not handled
    by test_multiple_formats_phasics_tif_ignore_h5
    and test_multiple_formats_phasics_tif
    should not be supported.
    """
    path, _files2 = setup_folder_single_holo()
    other_fmt_path = path / "single_holo.npy"
    np.save(other_fmt_path, np.zeros((10, 10)))
    try:
        qpformat.load_data(path)
    except qpformat.file_formats.MultipleFormatsNotSupportedError:
        pass
    else:
        assert False, "multiple formats should not be supported"


def test_series_format_holo_kw():
    """Siedband kwarg should be passed to subformats"""
    # combine a zip file with a regular hologram file
    path, _files2 = setup_folder_single_holo()
    bad_path = datapath / "series_phasics.zip"
    shutil.copy2(bad_path, path)
    ds1 = qpformat.load_data(path, holo_kw={"sideband": 1})
    ds2 = qpformat.load_data(path, holo_kw={"sideband": -1})
    p1 = ds1.get_qpimage(0)
    p2 = ds2.get_qpimage(0)
    assert np.all(p1.pha + p2.pha == 0)


def test_series_format_ignored():
    """Series file formats are ignored in SeriesFolder"""
    # combine a zip file with a regular hologram file
    path, _files2 = setup_folder_single_holo()
    bad_path = datapath / "series_phasics.zip"
    shutil.copy2(bad_path, path)
    ds = qpformat.load_data(path)
    assert len(ds) == 2
    assert bad_path not in ds.files


if __name__ == "__main__":
    test_multiple_formats_error()
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

import numpy as np
import qpimage

from .file_formats import formats, formats_dict, UnknownFileFormatError, \
    WrongFileFormatError


def guess_format(path):
    """Determine the file format of a folder or a file"""
    for fmt in formats:
        if fmt.verify(path):
            return fmt.__name__
    else:
        msg = "Undefined file format: '{}'".format(path)
        raise UnknownFileFormatError(msg)


def load_data(path, fmt=None, bg_data=None, bg_fmt=None,
              meta_data=None, holo_kw=None, as_type="float32"):
    """Load experimental data

    Parameters
    ----------
    path: str or pathlib.Path
        Path to experimental data file or folder
    fmt: str
        The file format to use (see `file_formats.formats`).
        If set to `None`, the file format is guessed.
    bg_data: str or pathlib.Path
        Path to background data file or `qpimage.QPImage`
    bg_fmt: str
        The file format to use (see `file_formats.formats`)
        for the background. If set to `None`, the file format
        is be guessed.
    meta_data: dict
        Meta data (see `qpimage.meta.META_KEYS`)
    holo_kw: dict
        Keyword arguments for hologram data; See
        :func:`qpimage.holo.get_field` for valid keyword arguments.
    as_type: str
        Defines the data type that the input data is casted to.
        The default is "float32" which saves memory. If high
        numerical accuracy is required (does not apply for a
        simple 2D phase analysis), set this to double precision
        ("float64").

    Returns
    -------
    dataobj: SeriesData or SingleData
        Object that gives lazy access to the experimental data.
    """
    if holo_kw is None:
        holo_kw = {}
    if meta_data is None:
        meta_data = {}
    path = pathlib.Path(path).resolve()
    # sanity checks
    for kk in meta_data:
        if kk not in qpimage.meta.META_KEYS:
            msg = "Meta data key not allowed: {}".format(kk)
            raise ValueError(msg)

    # ignore None or nan values in meta_data
    for kk in list(meta_data.keys()):
        if meta_data[kk] in [np.nan, None]:
            meta_data.pop(kk)

    if fmt is None:
        fmt = guess_format(path)
    else:
        if not formats_dict[fmt].verify(path):
            msg = "Wrong file format '{}' for '{}'!".format(fmt, path)
            raise WrongFileFormatError(msg)

    dataobj = formats_dict[fmt](path=path,
                                meta_data=meta_data,
                                holo_kw=holo_kw,
                                as_type=as_type)

    if bg_data is not None:
        if isinstance(bg_data, qpimage.QPImage):
            # qpimage instance
            dataobj.set_bg(bg_data)
        else:
            # actual data on disk
            bg_path = pathlib.Path(bg_data).resolve()
            if bg_fmt is None:
                bg_fmt = guess_format(bg_path)
                bgobj = formats_dict[bg_fmt](path=bg_path,
                                             meta_data=meta_data,
                                             holo_kw=holo_kw,
                                             as_type=as_type)
                dataobj.set_bg(bgobj)

    return dataobj



import functools
import warnings

import h5py
import numpy as np
import qpimage

from ..dataset import SeriesData


class NoSinogramDataFoundError(BaseException):
    pass


class SeriesHDF5GenericWarning(UserWarning):
    pass


class SeriesHDF5SinogramGeneric(SeriesData):
    """Base class for HDF5-based raw sinogram QPI data
    """
    priority = -9  # higher priority, because it's fast

    def __len__(self):
        return len(self._get_data_indices())

    @functools.lru_cache()
    def _get_data_indices(self):
        """Get all experiments from the hdf5 file"""
        with h5py.File(name=self.path, mode="r") as h5:
            if "sinogram" not in h5:
                raise NoSinogramDataFoundError(
                    f"Group 'sinogram' not found in '{self.path}'!")
            indices = sorted(h5["sinogram"].keys(), key=lambda x: int(x))

        if not indices:
            raise NoSinogramDataFoundError(
                f"No sinogram data found in '{self.path}'!")
        return indices

    def _get_metadata(self, dataset):
        """Return the metadata of the specified `h5py.Dataset`"""
        meta = qpimage.meta.MetaDict()
        for key in qpimage.META_KEYS:
            if key in dataset.attrs:
                meta[key] = dataset.attrs[key]
        return meta

    @staticmethod
    def verify(path):
        """Verify that `path` has generic sinogram data"""
        valid = False
        try:
            h5 = h5py.File(path, mode="r")
        except (OSError, IsADirectoryError):
            pass
        else:
            if ("file_format" in h5.attrs
                and "qpformat" in h5.attrs["file_format"].lower()
                    and "sinogram" in h5):
                valid = True
            h5.close()
        return valid

    @property
    @functools.lru_cache()
    def shape(self):
        with h5py.File(name=self.path, mode="r") as h5:
            group = h5["sinogram"]["0"]
            for key in ["field", "phase", "amplitude", "intensity"]:
                if key in group:
                    return len(self), group[key].shape[0], group[key].shape[1]
            else:
                # Fallback to expensive shape computation
                warnings.warn(f"Using fallback `shape` for '{self.path}'!",
                              SeriesHDF5GenericWarning)
                return super(SeriesHDF5SinogramGeneric, self).shape


class SeriesHDF5SinogramMeep(SeriesHDF5SinogramGeneric):
    """sinograms extracted from Meep/FDTD simulations

    I introduced this format in 2022 as part of my efforts to make
    the finite-difference time domain simulations from the ODTbrain
    manuscript :cite:`Mueller2015` publicly available.

    The HDF5 file contains a "background" and a "sinogram" group.
    The subgroups of "sinogram" are enumerated starting with "0".
    Each of them contain the complex "field" at a plane behind the
    scattering phantom as an HDF5 Dataset. The location of the plane
    (and all other relevant metadata) is stored in the attributes
    of this Dataset. In the same group, there are also the C++
    "simulation_code" and the log "simulation_output" which can
    be used to reproduce the simulation.
    """
    storage_type = "field"

    def __init__(self, path, meta_data=None, *args, **kwargs):
        """Initialize with default wavelength of 500nm"""
        if meta_data is None:
            meta_data = {}
        if "wavelength" not in meta_data:
            meta_data["wavelength"] = 500e-9
        super(SeriesHDF5SinogramMeep, self).__init__(
            path=path, meta_data=meta_data, *args, **kwargs)

        # set background data
        with h5py.File(path, "r") as h5:
            if "background" in h5:
                bgds = h5["background"]["field"]
                meta_data = self._get_metadata(bgds)
                qpi_bg = qpimage.QPImage(data=bgds[:],
                                         which_data="field",
                                         meta_data=meta_data,
                                         h5dtype=self.as_type)
                self.set_bg(qpi_bg)

    def _get_metadata(self, dataset):
        """Return simulation-specific metadata

        This uses the metadata previously extracted
        from the simulation and set in `dataset.attrs`
        to populate the QPI metadata.
        """
        meta = super(SeriesHDF5SinogramMeep, self)._get_metadata(dataset)
        meta.update(self.meta_data)
        if "ACQUISITION_PHI" in dataset.attrs:
            meta["angle"] = dataset.attrs["ACQUISITION_PHI"]

        if "MEDIUM_RI" in dataset.attrs:
            meta["medium index"] = dataset.attrs["MEDIUM_RI"]

        if "SAMPLING" in dataset.attrs:
            meta["pixel size"] = meta["wavelength"] / dataset.attrs["SAMPLING"]

            if "extraction focus distance [px]" in dataset.attrs:
                focus_px = dataset.attrs["extraction focus distance [px]"]
                meta["focus"] = focus_px * meta["pixel size"]

        meta["sim center"] = np.array(dataset.shape) / 2
        meta["sim model"] = "fdtd"

        return meta

    def get_qpimage_raw(self, idx=0):
        """Return QPImage without background correction"""
        name = self._get_data_indices()[idx]
        with h5py.File(name=self.path, mode="r") as h5:
            dataset = h5["sinogram"][name]["field"]
            meta_data = self._get_metadata(dataset)
            meta_data["time"] = float(idx)
            qpi = qpimage.QPImage(data=dataset[:],
                                  which_data="field",
                                  meta_data=meta_data,
                                  h5dtype=self.as_type)
        # set identifier
        qpi["identifier"] = self.get_identifier(idx)
        return qpi

    @staticmethod
    def verify(path):
        """Verify the file format

        The "file_format" attribute of the HDF5 file must contain
        the strings "qpformat" and "meep".
        """
        valid = SeriesHDF5SinogramGeneric.verify(path)
        if valid:
            with h5py.File(path, "r") as h5:
                valid = "meep" in h5.attrs["file_format"].lower()
        return valid



import pathlib

import numpy as np

import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_series_raw_oah():
    ds = qpformat.load_data(datapath / "series_hdf5_raw-oah.h5")
    assert ds.format == "SeriesHDF5RawOAH"
    assert len(ds) == 2
    qpi1 = ds.get_qpimage(0)
    qpi2 = ds.get_qpimage(1)
    assert qpi1.meta["wavelength"] == 532e-9
    assert qpi1.meta["numerical aperture"] == 1.0
    assert np.allclose(qpi1.meta["pos x"], -0.0002045580000000009)
    assert qpi1.meta["identifier"] == "a66ad:1"
    assert qpi1.meta["time"] == 2.5
    assert qpi2.meta["time"] == 2.8



from .series_hdf5_hyperspy import SeriesHdf5HyperSpy
from .series_hdf5_raw_oah import SeriesHDF5RawOAH
from .series_zip_tif_holo import SeriesZipTifHolo
from .single_hdf5_raw_oah import SingleHDF5RawOAH
from .single_tif_holo import SingleTifHolo


registered_formats = [
    SeriesHDF5RawOAH,
    SeriesHdf5HyperSpy,
    SeriesZipTifHolo,
    SingleHDF5RawOAH,
    SingleTifHolo,
]



import pathlib
import tempfile

import numpy as np
import qpimage

import qpformat.core


datapath = pathlib.Path(__file__).parent / "data"


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path = tempfile.mktemp(suffix=".h5", prefix="drymass_test_convert")
    dout = tempfile.mkdtemp(prefix="drymass_test_sphere_")
    with qpimage.QPSeries(h5file=path, h5mode="w") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    return qpi, path, dout


def test_change_wavelength():
    _qpi, path, dout = setup_test_data()

    ds1 = qpformat.load_data(path, meta_data={"wavelength": 500e-9})
    ds2 = qpformat.load_data(path, meta_data={"wavelength": 333e-9})

    assert ds1.identifier != ds2.identifier, "should be different identifiers"


def test_identifier():
    path = datapath / "series_phasics.zip"

    ds1 = qpformat.load_data(path=path)
    assert ds1.identifier == "49bbc"

    bg_data = ds1.get_qpimage(0)
    ds2 = qpformat.load_data(path=path, bg_data=bg_data)
    assert ds2.background_identifier == "bf13f"
    assert ds2.identifier == "b6d1c"


def test_meta():
    data = np.ones((20, 20), dtype=float)
    tf = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(tf, data)

    ds = qpformat.load_data(path=tf, meta_data={"time": 47})
    assert ds.get_time() == 47
    # use `.name` because of short-hand paths on Windows
    assert pathlib.Path(tf).name in ds.get_name()


def test_meta_none():
    data = np.ones((20, 20), dtype=float)
    tf = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(tf, data)

    ds = qpformat.load_data(path=tf, meta_data={"time": 47,
                                                "medium index": None,
                                                "wavelength": np.nan})

    assert "time" in ds.meta_data
    assert "medium index" not in ds.meta_data
    assert "wavelength" not in ds.meta_data


def test_repr():
    data = np.ones((20, 20), dtype=float)
    tf = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(tf, data)

    ds1 = qpformat.load_data(path=tf, meta_data={"wavelength": 500e-9,
                                                 "pixel size": .15e-6})
    assert "λ=500" in ds1.__repr__()
    assert "1px=0.15" in ds1.__repr__()

    ds2 = qpformat.load_data(path=tf, meta_data={"wavelength": 1e-9,
                                                 "pixel size": 15})
    assert "λ=1" in ds2.__repr__()
    assert "1px=15" in ds2.__repr__()


def test_save():
    data_dir = tempfile.mkdtemp(prefix="qpformat_test_data_")
    data_dir = pathlib.Path(data_dir)

    data1 = np.ones((20, 20), dtype=float)
    data1 *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    np.save(data_dir / "data1.npy", data1)

    data2 = data1 * np.linspace(1.33, 1.04, 20).reshape(1, -1)
    np.save(data_dir / "data2.npy", data2)

    # save as h5
    ds = qpformat.load_data(path=data_dir)
    save_dir = tempfile.mkdtemp(prefix="qpformat_test_data_save_")
    save_path = pathlib.Path(save_dir) / "savetest.h5"
    ds.saveh5(save_path)

    # load h5
    ds2 = qpformat.load_data(path=save_path)

    assert ds.identifier is not None
    assert len(ds) == len(ds2)
    assert np.all(ds.get_qpimage(0).pha == ds2.get_qpimage(0).pha)


def test_save_one_bg():
    """Test that if there is a single bg, it is stored somehow
    in the output file"""
    data_dir = tempfile.mkdtemp(prefix="qpformat_test_data_")
    data_dir = pathlib.Path(data_dir)

    data1 = np.ones((20, 20), dtype=float)
    data1 *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    np.save(data_dir / "data1.npy", data1)

    data2 = data1 * np.linspace(1.33, 1.04, 20).reshape(1, -1)
    np.save(data_dir / "data2.npy", data2)

    # save as h5
    bg = qpimage.QPImage(data=data1, which_data="phase")
    ds = qpformat.load_data(path=data_dir, bg_data=bg)
    save_dir = tempfile.mkdtemp(prefix="qpformat_test_data_save_")
    save_path = pathlib.Path(save_dir) / "savetest.h5"
    ds.saveh5(save_path)

    # load h5
    ds2 = qpformat.load_data(path=save_path)
    assert np.all(ds.get_qpimage(0).pha == ds2.get_qpimage(0).pha)
    assert np.all(ds2.get_qpimage(0).pha == 0)
    assert not np.all(ds2.get_qpimage(1).pha == 0)


def test_save_identifier():
    """Test that identifier is saved only when full series was saved"""
    data_dir = tempfile.mkdtemp(prefix="qpformat_test_data_")
    data_dir = pathlib.Path(data_dir)

    data1 = np.ones((20, 20), dtype=float)
    data1 *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    np.save(data_dir / "data1.npy", data1)

    data2 = data1 * np.linspace(1.33, 1.04, 20).reshape(1, -1)
    np.save(data_dir / "data2.npy", data2)

    # save as h5
    bg = qpimage.QPImage(data=data1, which_data="phase")
    ds = qpformat.load_data(path=data_dir, bg_data=bg)
    save_dir = tempfile.mkdtemp(prefix="qpformat_test_data_save_")
    save_path1 = pathlib.Path(save_dir) / "savetest1.h5"
    save_path2 = pathlib.Path(save_dir) / "savetest2.h5"
    ds.saveh5(save_path1)
    ds.saveh5(save_path2, qpi_slice=(slice(0, 10), slice(0, 10)))

    # load h5
    with qpimage.QPSeries(h5file=save_path1) as qps1:
        assert qps1.identifier is not None
    with qpimage.QPSeries(h5file=save_path2) as qps2:
        assert qps2.identifier is None


def test_set_bg():
    data = np.ones((20, 20), dtype=float)
    data *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    f_data = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(f_data, data)

    bg_data = np.ones((20, 20), dtype=float)
    bg_data *= np.linspace(0, 1.1, 20).reshape(1, -1)
    f_bg_data = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(f_bg_data, bg_data)

    # set bg with dataset
    ds1 = qpformat.core.load_data(path=f_data)
    bg1 = qpformat.core.load_data(path=f_bg_data)
    ds1.set_bg(bg1)
    assert np.allclose(ds1.get_qpimage().pha, data - bg_data)

    # set with QPImage
    ds2 = qpformat.core.load_data(path=f_data)
    bg2 = qpformat.core.load_data(path=f_bg_data)
    ds2.set_bg(bg2.get_qpimage())
    assert np.allclose(ds2.get_qpimage().pha, data - bg_data)

    # set with list
    ds3 = qpformat.core.load_data(path=f_data)
    bg3 = qpformat.core.load_data(path=f_bg_data)
    ds3.set_bg([bg3.get_qpimage()])
    assert np.allclose(ds3.get_qpimage().pha, data - bg_data)

    # set with bad value
    ds4 = qpformat.core.load_data(path=f_data)
    try:
        ds4.set_bg(data)
    except ValueError:
        pass
    else:
        assert False, "ndarray cannot be used to set bg data"


def test_set_bg_series():
    data_dir = tempfile.mkdtemp(prefix="qpformat_test_data_")
    data_dir = pathlib.Path(data_dir)

    data1 = np.ones((20, 20), dtype=float)
    data1 *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    np.save(data_dir / "data1.npy", data1)

    data2 = data1 * np.linspace(1.33, 1.04, 20).reshape(1, -1)
    np.save(data_dir / "data2.npy", data2)

    # bg data
    bg_data_dir = tempfile.mkdtemp(prefix="qpformat_test_bg_data_")
    bg_data_dir = pathlib.Path(bg_data_dir)

    bg_data1 = data1 * np.linspace(1.0, 1.02, 20).reshape(1, -1)
    np.save(bg_data_dir / "bg_data1.npy", bg_data1)

    bg_data2 = data1 * np.linspace(.9, 0.87, 20).reshape(-1, 1)
    np.save(bg_data_dir / "bg_data2.npy", bg_data2)

    # tests
    ds1 = qpformat.core.load_data(path=data_dir, as_type="float64")
    bg1 = qpformat.core.load_data(path=bg_data_dir, as_type="float64")
    assert len(ds1) == 2
    ds1.set_bg(bg1)

    assert np.allclose(ds1.get_qpimage(0).pha, data1 - bg_data1)
    assert np.allclose(ds1.get_qpimage(1).pha, data2 - bg_data2)
    assert not np.allclose(ds1.get_qpimage(0).pha, data2 - bg_data2)


def test_set_bg_qpimage():
    data = np.ones((20, 20), dtype=float)
    data *= np.linspace(-.1, 3, 20).reshape(-1, 1)
    f_data = tempfile.mktemp(prefix="qpformat_test_", suffix=".npy")
    np.save(f_data, data)

    bg_data = np.ones((20, 20), dtype=float)
    bg_data *= np.linspace(0, 1.1, 20).reshape(1, -1)
    qpi = qpimage.QPImage(data=bg_data, which_data="phase")

    # set bg with dataset
    ds1 = qpformat.core.load_data(path=f_data, bg_data=qpi)
    assert np.allclose(ds1.get_qpimage().pha, data - bg_data)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import tempfile

import h5py
import numpy as np

import qpimage
import qpformat


datapath = pathlib.Path(__file__).parent / "data"


def test_identifier():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi1 = qpimage.QPImage(h5file=path, h5mode="r")
    qpi2 = qpi1.copy()
    qpi2["identifier"] = "an important string"
    # generate qpseries hdf5 file
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"):
        pass

    ds = qpformat.load_data(tf)
    # individual identifiers are not extracted anymore as of
    # qpformat version 0.3.4
    assert ds.get_identifier(0) != "an important string"
    assert ds.get_identifier(1) != "an important string"
    assert ds.identifier in ds.get_identifier(0)


def test_load_data():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi1 = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi2 = qpi1.copy()
    qpi1["identifier"] = "test100"
    qpi2["identifier"] = "test200"
    # generate qpseries hdf5 file
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"):
        pass

    ds = qpformat.load_data(tf)
    assert len(ds) == 2
    assert ds.path.samefile(pathlib.Path(tf))
    assert np.isnan(ds.get_time(1))
    assert "SeriesHdf5Qpimage" in ds.__repr__()
    qpd = ds.get_qpimage(1)
    assert qpd["identifier"] != qpi1["identifier"]
    assert qpd == qpi1
    assert qpd.shape == qpi1.shape
    assert np.allclose(qpd.amp, qpi1.amp)
    assert np.allclose(qpd.pha, qpi1.pha)


def test_meta_extraction():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r")
    # generate qpseries hdf5 file
    with qpimage.QPSeries(qpimage_list=[qpi, qpi],
                          h5file=tf,
                          h5mode="a",
                          meta_data={"wavelength": 111e-9,
                                     "pixel size": .12}):
        pass

    ds = qpformat.load_data(tf)

    assert ds.meta_data["wavelength"] == 111e-9
    assert ds.meta_data["pixel size"] == .12


def test_meta_override():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r")
    # generate qpseries hdf5 file
    with qpimage.QPSeries(qpimage_list=[qpi, qpi],
                          h5file=tf,
                          h5mode="a",
                          meta_data={"wavelength": 111e-9,
                                     "pixel size": .12}):
        pass

    wl = 333e-9
    px = .111
    ds = qpformat.load_data(tf, meta_data={"wavelength": wl,
                                           "pixel size": px})

    assert ds.meta_data["wavelength"] == wl
    assert ds.meta_data["pixel size"] == px

    qpi_ds = ds.get_qpimage(0)
    assert qpi_ds["wavelength"] == wl
    assert qpi_ds["pixel size"] == px

    qpi_ds = ds.get_qpimage(1)
    assert qpi_ds["wavelength"] == wl
    assert qpi_ds["pixel size"] == px


def test_returned_identifier():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r")
    # generate qpseries hdf5 file
    with qpimage.QPSeries(qpimage_list=[qpi, qpi],
                          h5file=tf,
                          h5mode="a"):
        pass

    ds = qpformat.load_data(tf)

    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


def test_subjoined_load_data():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_subjoined")
    qpi1 = qpimage.QPImage(h5file=path, h5mode="r").copy()
    qpi2 = qpi1.copy()
    qpi1["identifier"] = "test100"
    qpi2["identifier"] = "test200"
    # generate subjoined qpseries hdf5 file
    with h5py.File(tf, mode="w") as h5:
        qps = h5.require_group("qpseries")
        qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                         h5file=qps)

    ds = qpformat.load_data(tf)
    assert len(ds) == 2
    assert ds.path.samefile(pathlib.Path(tf))
    assert np.isnan(ds.get_time(1))
    assert "SeriesHdf5QpimageSubjoined" in ds.__repr__()
    qpd = ds.get_qpimage(1)
    assert qpd["identifier"] != qpi1["identifier"]
    assert qpd == qpi1
    assert qpd.shape == qpi1.shape
    assert np.allclose(qpd.amp, qpi1.amp)
    assert np.allclose(qpd.pha, qpi1.pha)


def test_subjoined_meta_extraction():
    path = datapath / "single_qpimage.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpformat_test_")
    qpi = qpimage.QPImage(h5file=path, h5mode="r")

    # generate subjoined qpseries hdf5 file
    with h5py.File(tf, mode="w") as h5:
        qps = h5.require_group("qpseries")
        qpimage.QPSeries(qpimage_list=[qpi, qpi],
                         h5file=qps,
                         meta_data={"wavelength": 111e-9,
                                    "pixel size": .12})

    ds = qpformat.load_data(tf)

    assert ds.meta_data["wavelength"] == 111e-9
    assert ds.meta_data["pixel size"] == .12


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'library for opening quantitative phase imaging data'
name = 'qpformat'
year = "2017"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version  # noqa: E402

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/RI-imaging/qpformat',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["h5py>=2.7.0",
                      "numpy>=1.12.0",
                      "qpimage>=0.7.7",
                      "tifffile>=2020.5.25",
                      ],
    python_requires='>=3.9, <4',
    entry_points={
       "console_scripts": [
           "qpinfo = qpformat.cli:qpinfo",
            ],
       },
    keywords=["data file format",
              "digital holographic microscopy",
              "quantitative phase imaging",
              ],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )



import os
import pathlib
import tempfile
import warnings

import h5py
import tifffile

import qpformat
from qpformat.file_formats import WrongFileFormatError
from qpformat.file_formats.fmts_raw_oah.series_hdf5_hyperspy import (
    HyperSpyNoDataFoundError, WrongSignalTypeWarnging)

datapath = pathlib.Path(__file__).parent / "data"


def make_hyperspy(signal_type="hologram"):
    """Create a basic hyperspy file"""
    tifin = datapath / "single_holo.tif"
    with tifffile.TiffFile(os.fspath(tifin)) as tf:
        data = tf.pages[0].asarray()
    tdir = pathlib.Path(tempfile.mkdtemp(prefix="qpformat_test_hyperspy"))
    hspyf = tdir / "test.h5"
    with h5py.File(hspyf, mode="w") as h5:
        h5.attrs["file_format"] = "hyperspy"
        exp = h5.create_group("Experiments")
        hol = exp.create_group("Hologram of an HL60 cell")
        hol.create_dataset(name="data", data=data)
        # signal
        sig = hol.create_group("metadata/Signal")
        sig.attrs["signal_type"] = signal_type
        # set pixel size
        for name in ["axis-0", "axis-1"]:
            axi = hol.create_group(name)
            axi.attrs["scale"] = 107
            axi.attrs["units"] = "nm"

    return tdir, hspyf


def test_basic():
    tdir, hspyf = make_hyperspy()
    ds = qpformat.load_data(hspyf)
    qpi = ds.get_qpimage(0)

    assert len(ds) == 1
    assert qpi["pixel size"] == 0.107e-6


def test_returned_identifier():
    tdir, hspyf = make_hyperspy()
    ds = qpformat.load_data(hspyf)
    qpi = ds.get_qpimage(0)
    assert "identifier" in qpi
    qpiraw = ds.get_qpimage_raw(0)
    assert "identifier" in qpiraw


def test_wrong_format():
    path = datapath / "single_qpimage.h5"
    try:
        qpformat.load_data(path, fmt="SeriesHdf5HyperSpy")
    except WrongFileFormatError:
        pass
    else:
        raise ValueError("qpimage data is not hyperspy data")


def test_wrong_signal_type():
    tdir, hspyf = make_hyperspy(signal_type="unknown")
    ds = qpformat.load_data(hspyf)

    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        try:
            ds.get_qpimage(0)
        except HyperSpyNoDataFoundError:
            pass
        else:
            raise ValueError("Error due to wrong signal type")

        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, WrongSignalTypeWarnging)
        assert "unknown" in str(w[-1].message)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



r"""Conversion of external file formats to .npy files

Sometimes the data recorded are not in a file format supported
by qpformat or it is not feasible to implement a reader
class for a very unique data set. In this example, QPI data,
stored as a tuple of files ("\*_intensity.txt" and "\*_phase.txt")
with commas as decimal separators, are converted to the numpy
file format which is supported by qpformat.

This example must be executed with a directory as an
command line argument, i.e.
``python convert_txt2npy.py /path/to/folder/``
"""
import pathlib
import sys

import numpy as np


def get_paths(folder):
    '''Return *_phase.txt files in `folder`'''
    folder = pathlib.Path(folder).resolve()
    files = folder.rglob("*_phase.txt")
    return sorted(files)


def load_file(path):
    '''Load a txt data file'''
    path = pathlib.Path(path)
    data = path.open().readlines()
    # remove comments and empty lines
    data = [ll for ll in data if len(ll.strip()) and not ll.startswith("#")]
    # determine data shape
    n = len(data)
    m = len(data[0].strip().split())
    res = np.zeros((n, m), dtype=np.dtype(float))
    # write data to array, replacing comma with point decimal separator
    for ii in range(n):
        res[ii] = np.array(data[ii].strip().replace(",", ".").split(),
                           dtype=float)
    return res


def load_field(path):
    '''Load QPI data using *_phase.txt files'''
    path = pathlib.Path(path)
    phase = load_file(path)
    inten = load_file(path.parent / (path.name[:-10] + "_intensity.txt"))
    ampli = np.sqrt(inten)
    return ampli * np.exp(1j * phase)


if __name__ == "__main__":
    path = pathlib.Path(sys.argv[-1])
    if not path.is_dir():
        raise ValueError("Command line argument must be directory!")
    # output directory
    pout = path.parent / (path.name + "_npy")
    pout.mkdir(exist_ok=True)
    # get input *_phase.txt files
    files = get_paths(path)
    # conversion
    for ff in files:
        field = load_field(ff)
        np.save(str(pout / (ff.name[:-10] + ".npy")), field)



"""HyperSpy hologram file format

This example demonstrates the import of hologram images in
the `HyperSpy <https://hyperspy.org/>`_ hdf5 file format.
The off-axis electron hologram shows an electrically biased Fe
needle :cite:`Migunov2015`. The corresponding HyperSpy demo
can be found `here <http://nbviewer.jupyter.org/github/hyperspy/hypersp
y-demos/blob/master/electron_microscopy/Holography/Holography.ipynb>`_.
"""

import urllib.request
import os

import matplotlib.pylab as plt
import qpformat

# load the experimental data
dl_loc = "https://github.com/hyperspy/hyperspy/raw/RELEASE_next_major/" \
         + "hyperspy/misc/holography/example_signals/"
dl_name = "01_holo_Vbp_130V_0V_bin2_crop.hdf5"
if not os.path.exists(dl_name):
    print("Downloading {} ...".format(dl_name))
    urllib.request.urlretrieve(dl_loc + dl_name, dl_name)

ds = qpformat.load_data(dl_name,
                        holo_kw={
                            # reduces ringing artifacts in the amplitude image
                            "filter_name": "smooth disk",
                            # select correct sideband
                            "sideband": -1,
                            })

# retrieve the qpimage.QPImage instance
qpi = ds.get_qpimage(0)

# plot results
plt.figure(figsize=(8, 3.5))
# ampltitude
ax1 = plt.subplot(121, title="amplitude")
map1 = plt.imshow(qpi.amp, cmap="gray")
plt.colorbar(map1, ax=ax1, fraction=.0455, pad=0.04)
# phase in interval [-1rad, 1rad]
ax2 = plt.subplot(122, title="phase [rad]")
map2 = plt.imshow(qpi.pha)
plt.colorbar(map2, ax=ax2, fraction=.0455, pad=0.04)
# disable axes
[ax.axis("off") for ax in [ax1, ax2]]
plt.tight_layout()
plt.show()
